/* -*- mode: C++; c-file-style: "gnu" -*- */
/* argument.h: Qt wrapper for D-BUS message arguments
 *
 * Copyright (C) 2005  James Thorniley <james@uncommonlygood.co.uk>
 *
 * Licensed under the Academic Free License version 2.1
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


#ifndef DBUS_QT_ARGUMENT_H
#define DBUS_QT_ARGUMENT_H

#include "dbus/dbus.h"
#include <qglobal.h>

#ifdef __cplusplus //prevents Doxygen from deciding that Argument is a template class
template <class T> class QValueList;
#endif

class QStringList;
class QString;

namespace DBusQt
{

  /**
  
  @short A class for wrapping C++/Qt types for use with D-BUS
  
  <p>Each Argument object represents a <em>single complete type</em> as it is
  defined in the D-BUS protocol. That is, an argument can
  be a byte, boolean, integer, string etc. Or it could be a
  complex type such as an array or struct.
  
  <p>To attach data to messages, it must first be turned into Argument objects.
  Use one of the supplied constructors to do that. Add it to a message by getting
  hold of the message's argument list (see: {@link Message::arguments()}), and using
  QValueList methods to add your argument to the list.
  
  <p>For example to add a QString to a message, you might use:
  
  @code
  Message myMessage(...); //fill in arguments with destination, object etc...
  
  Argument stringArg( QString( "Hello world" ) );
  
  myMessage.arguments()->append( stringArg );
  
  @endcode
  
  <p>To get data out of a message, once again use the {@link Message::arguments() arguments} method to retreive
  a QValueList of arguments and retrieve an argument from the list. To convert this argument
  back to a C++ / Qt type, use the relevant toXXX() method. You can check which type
  this argument represents with the {@link #argumentType()} method.
  
  <p>For example to receive a reply containing a QString:
  
  @code
  Message reply = connection.sendWidthReplyAndBlock( myMessage );
  
  Argument replyArg = reply.arguments()->first();
  
  QString replyString = replyArg.toString();
  @endcode
  
  <p>You may wish to use the QValueList::Iterator class to iterate over multiple arguments.
  See the Qt documentation for more help.
  
  @todo Implement Dicts.
  
  @author James Thorniley
  */
  class Argument
  {

     
    public:

      /**
      The possible types that can be contained
      in an argument object.
      
      <p>The following table describes all the types you may
      use for D-BUS arguments.
      
      <table>
      <tr><td>Type</td><td>D-BUS format</td><td>C++/QT format</td></tr>
      
      <tr><td>Byte</td><td>Single byte</td><td>char</td></tr>
      <tr><td>Boolean</td><td>32 bit integer, 1 = true, 0 = false, other values invalid </td><td>bool</td></tr>
      <tr><td>Int16</td><td>16 bit signed integer</td><td>Q_INT16</td></tr>
      <tr><td>Uint16</td><td>16 bit unsigned integer</td><td>Q_UINT16</td></tr>
      <tr><td>Int32</td><td>32 bit signed integer</td><td>Q_INT32</td></tr>
      <tr><td>Uint32</td><td>32 bit unsigned integer</td><td>Q_UINT32</td></tr>
      <tr><td>Int64</td><td>64 bit signed integer</td><td>Q_LONG</td></tr>
      <tr><td>Uint64</td><td>64 bit unsigned integer</td><td>Q_ULONG</td></tr>
      <tr><td>IEEEFloat</td><td>IEEE 754 double precision float</td><td>double</td></tr>
      <tr><td>String</td><td>UTF-8 encoded null terminated byte array</td><td>QString</td></tr>
      <tr><td>Struct</td><td>Struct</td><td>QValueList&lt;Argument&gt;</tr>
      <tr><td>Invalid</td><td colspan="2">Arguments that do not represent any real type are of type Invalid.</td></tr>
      
      
      </table>
      */
      enum Type
      {
        Byte = (int)'y',
        Boolean = (int)'b',
        Int16 = (int)'n',
        Uint16 = (int)'q',
        Int32 = (int)'i',
        Uint32 = (int)'u',
        Int64 = (int)'x',
        Uint64 = (int)'t',
        IEEEFloat = (int)'d',
        String = (int)'s',
        Struct = (int)'r',
        Array = (int)'a',
        Invalid = 0 /** 
                <p>Does not represent a real type. Probably the {@link #argumentType()} of a
                object created with {@link #Argument() }, or the {@link #elementType()} of an
                Argument that is not an array.
                */
      };
     
      /**
      The possible types that may be nested.
      
      @see Argument( const QValueList<Argument>&, NestedType )
      */
      enum NestedType
      {
        ///An array of arrays.
        ArrayArray,
        ///A struct.
        Structure
      };
      
      /**
      Create an empty argument.
      
      <p>The new object represents a placeholder for an
      unknown / unsupported type.
      */
      Argument( );
      
      /**
      Create a new argument representing a 
      single byte.
      
      <p>This is equivalent to a DBUS_TYPE_BYTE (code 'y')
      on the wire.
      
      @param byte The byte to use in the argument
      @todo Implement Argument( char byte );
      */
      Argument( char byte );

      /**
      Create a new argument representing a
      boolean value.
      
      <p>This is equivalent to a DBUS_TYPE_BOOLEAN (code 'b')
      on the wire.
      
      @param boolean The boolean to use in the argument
      @todo Implement Argument( bool boolean );
      */
      Argument( bool boolean );

      /**
      Create a new argument representing a
      16 bit signed integer.
      
      <p>This is equivalent to a DBUS_TYPE_INT16 (code 'n')
      on the wire.
      
      @param integer The integer to use in the argument
      @todo Implement Argument( Q_INT16 integer );
      */      
      Argument( Q_INT16 integer );
      
      /**
      Create a new argument representing an
      unsigned 16 bit integer.
      
      <p>This is equivalent to a DBUS_TYPE_UINT16 (code 'q')
      on the wire.
      
      @param integer The integer to use in the argument
      @todo Implement Argument( Q_UINT16 integer );
      */
      Argument( Q_UINT16 integer );

      /**
      Create a new argument representing a
      signed 32 bit integer.
      
      <p>This is equivalent to a DBUS_TYPE_INT32 (code 'i')
      on the wire.
      
      @param integer The integer to use in the argument
      */      
      Argument( Q_INT32 integer );
      
      /**
      Create a new argument representing an
      unsigned 32 bit integer.
      
      <p>This is equivalent to a DBUS_TYPE_UINT32 (code 'u')
      on the wire.
      
      @param integer The integer to use in the argument
      @todo Implement Argument( Q_UINT32 integer );
      
      */
      Argument( Q_UINT32 integer );
      
      /**
      Create a new argument representing a
      signed 64 bit integer.
      
      <p>This is equivalent to a DBUS_TYPE_INT64 (code 'x')
      on the wire.
      
      @param integer The integer to use in the argument
      @todo Implement Argument( Q_LONG integer );
      */
      Argument( Q_LONG integer );

      /**
      @brief Create a new argument representing an unsigned 64 bit integer.
      
      <p>This is equivalent to a DBUS_TYPE_UINT64 (code 't')
      on the wire.
      
      @param integer The integer to use in the argument
      @todo Implement Argument( Q_ULONG integer );
      */ 
      Argument( Q_ULONG integer );
      
      /**
      @brief Create a new argument representing a IEEE-754 floating point number.
      
      <p>This is equivalent to a DBUS_TYPE_DOUBLE (code 'd')
      on the wire
      
      @param ieeeFloat The double to use in the argument
      @todo Implement Argument( doubel ieeeFloat );
      */
      Argument( double ieeeFloat );
      
      /**
      Create a new argument representing a UTF-8 string.
      
      <p>This is equivalent to a DBUS_TYPE_STRING (code 's')
      on the wire
      
      @param string The string to use in the argument
      */      
      Argument( const QString& string );
      
      /*!
      @brief Create a new argument representing a nested type. 
      This means either a struct or an array of arrays at the moment.
      
      <p>A struct, in D-BUS terms, may contain any number of
      single complete types, including other structs. This class
      models a struct as a QValueList of Argument objects.
      
      <p>To create an array of arrays,
      all the arguments in the given QValueList must be
      arrays (i.e. argumentType() == Array). The elementType of
      each of the arrays must be equal to the supplied element type.
      This element type can be Array as well, if you want to make
      and array of arrays of arrays.
      
      @param nested A QValueList of Argument objects, representing
      a struct or array array.
      @param type Either @arg @c Argument::ArrayArray for an array of arrays;
      or @arg @c Argument::Structure for a struct.
      
      @see #NestedType
      */
      Argument( const QValueList<Argument>& nested, NestedType type );
      
      /**
      Create a new argument representing an array of strings.
      
      @param stringList The array of strings to use.
      */
      Argument( const QStringList& stringList );
      
      /**
      Deep copy an Argument object.
      
      @param arg The Argument to copy.
      */
      Argument( const Argument &arg );
      
      /**
      Deep copy an Argument object.
      
      @param arg The Argument to copy.
      @return A copy of the supplied Argument.
      */
      Argument &operator=( const Argument &arg );
      
      /**
      Destructor.
      */
      ~Argument();
      
      /**
      Gets the type of this argument.
      
      @return The type of this argument.
      
      @see #Type
      */
      Type argumentType() const ;
      
      /**
      Gets the type of the elements of this array
      (if this argument is an array)
     
      @return The type of this array
      
      @see #Type
      */
      Type elementType() const;
      
      /**
      Converts the argument to a QString. If the argument
      is not a string, returns QString::null
      
      @return The QString held in this argument, or QString::null
      if this argument is not a string
      */
      const QString toString() const;
      
      /**
      Converts the argument to a QStringList, assuming the
      argument is an array of strigs.
      
      @return The QStringList held in this argument, or an
      invalid QStringList if this argument is not a string array.
      */
      const QStringList toStringList() const;
      
      /**
      Converts this argument into a QValueList of arguments
      representing a struct object, assuming this argument
      is actually a struct.
      
      @return The QValueList held in this argument, or an
      invalid QValueList&lt;Argument&gt; if this argument
      is not a struct.
      */
      const QValueList<Argument> &toStruct() const;
      
      /**
      Converts this argument into a QValueList of arguments
      representing an array or arrays. Each argument in the
      list will be themselves arrays, and they will all have
      the same element type.
      
      @return The QValueList held in this argument, or
      an invalid QValueList&lt;Argument&gt; if this argument
      is not an array of arrays.
      */
      const QValueList<Argument> &toArrayArray() const;
      
      Q_INT32 toInt32() const;
      
    private:
    
      class Private;
      Private *d;
    
  };
  
}

#endif
